*
* 8031/51 Startup Code & Runtime library for COMPACT model
*
* Copyright 1991,1992 Dave Dunfield
* All rights reserved.
*
* System memory map
?RAM  EQU $3000   RAM Starts here
?RAMEND EQU $3FFF   RAM Ends here
  ORG $2200   ROM Starts here
* Fixed memory locations for alternate access to the CPU register bank.
* If you are NOT useing BANK 0, these equates must be adjusted.
?R0	EQU	0		Used for "POP" from stack
?R1	EQU	?R0+1		Used to load index indirectly
?R2	EQU	?R0+2		""		""		""		""
?R3	EQU	?R0+3		Used by some runtime lib functions
?R4	EQU	?R0+4
?R5	EQU	?R0+5
?R6	EQU	?R0+6
?R7	EQU	?R0+7
*
* Startup code entry point
*
* If you are NOT using interrupts, you can reclaim 50 bytes
* of code space by removing the following TWO lines.
	AJMP	*+$0032		Skip interrupt vectors
	DS	$0032-2		Reserve space for interrupt vectors
*
	MOV	SP,#?stk-1	Set up initial stack
	MOV	?stack,#?RAMEND	Set LOW byte of internal stack
	MOV	?stack+1,#=?RAMEND	
	MOV	DPTR,#?heap	Point to heap storage
	CLR	A		Get end of list flag
	MOVX	[DPTR],A	Set heap ending address
	LCALL	main		Execute program
* EXIT to MON51 by calling the 'timer1' interrupt vector.
* This causes MON51 to think that a single-step operation has just
* completed, and therefore it saves the user registers, and performs
* a context switch back to the monitor.
exit	LCALL	$001B		Call Timer-1 interrupt
	SJMP	exit		Incase he go's again
*
* Compiler support routines
*
* Setup R0 to access the stack
?auto0	XCH	A,R0		Save A, get parameter
	ADD	A,SP		Adjust for stack offset
	XCH	A,R0		And restore order
	RET
* Setup R1 to access the stack
?auto1	XCH	A,R1		Save A, get parmater
	ADD	A,SP		Adjust for stack offset
	XCH	A,R1		And restore order
	MOV	R2,#0		Clear high of index
	RET
* Adjust the stack pointer by R7 bytes
?adjstk	POP	DPH		Get high byte of RET addr
	POP	DPL		Get low byte of RET addr
	XCH	A,R7		Get adjust (and save A)
	ADD	A,SP		Adjust stack
	MOV	SP,A		Set new value
	XCH	A,R7		Restore A (& return new stack)
	PUSH	DPL		Resave low
	PUSH	DPH		Resave high
	RET
* Clean up stack & exit function
?exit	XCH	A,R7		Get adjust (save A)
	ADD	A,SP		Adjust stack
	MOV	SP,A		Resave
	XCH	A,R7		Restore A (& return new stack)
	RET
* Load a byte from external memory into TEMP register
?extb	MOV	R3,A		Save A
	MOVX	A,[DPTR]	Get the byte
	XCH	A,R3		Restore A, save result
	RET
* Load a word of external memory into TEMP register
?extw	MOV	R4,A		Save A
	MOVX	A,[DPTR]	Get the low order byte
	MOV	R3,A		Save result LOW
	INC	DPTR		Advance to next
	MOVX	A,[DPTR]	Get the high order byte
	XCH	A,R4		Restore A, save result HIGH
	RET
* Load a word of external memory into INDEX register
?exti	MOV	R2,A		Save A
	MOVX	A,[DPTR]	Get the low order byte
	MOV	R1,A		Save result LOW
	INC	DPTR		Advance to next
	MOVX	A,[DPTR]	Get the high order byte
	XCH	A,R2		Restore A, Save result HIGH
	RET
* Add accumulator to INDEX register
?addi	ADD	A,R1		Add low bytes
	MOV	R1,A		Resave
	MOV	A,B		Get high byte
	ADDC	A,R2		Add hich bytes
	MOV	R2,A		Resave
	RET
* Sign extend TEMP register
?tsign	MOV	R4,#0		Assume positive
	XCH	A,R3		Get low byte
	JNB	A.7,*+4		Its negative
	DEC	R4		Convert to negative
	XCH	A,R3		Put it back
	RET
* Sign extend A into B
?sign	MOV	B,#0		Assume positive
	JNB	A.7,*+5		Assumption correct
	DEC	B		Convert to -1
	RET
*
* Perform a SWITCH statement
*
?switch	MOV	R5,A		Save low VALUE
* Examine this switch table entry
?swit1	CLR	A		Offset to LOW label
	MOVC	A,[A+DPTR]	Get LOW byte of label
	MOV	R7,A		Save for later
	MOV	A,#1		Offset to HIGH label
	MOVC	A,[A+DPTR]	Get HIGH of label
	ORL	A,R7		At end of table?
	JNZ	?swit2		No, keep looking
* End of table, advance to default & jump
	INC	DPTR		Advance
	INC	DPTR		Advance
	SJMP	?swit3		And proceed
* Not end of table, check the value
?swit2	MOV	A,#2		Offset to LOW value
	MOVC	A,[A+DPTR]	Get LOW byte of value
	CJNE	A,?R5,?swit4	Dosn't match, skip
	MOV	A,#3		Offset to HIGH value
	MOVC	A,[A+DPTR]	Get HIGH byte of value
	CJNE	A,B,?swit4	Dosn't match, skip
* Execute this routine
?swit3	CLR	A		Zero offset
	MOVC	A,[A+DPTR]	Get LOW byte of label
	PUSH	A		Stack for RET
	MOV	A,#1		Offset to HIGH
	MOVC	A,[A+DPTR]	Get HIGH byte of label
	PUSH	A		Stack for RET
	RET			Execute the code
* No match, loop and test next entry
?swit4	INC	DPTR		Skip LOW label
	INC	DPTR		Skip HIGH label
	INC	DPTR		Skip LOW value
	INC	DPTR		Skip HIGH value
	SJMP	?swit1		And keep looking
*
* 16 bit multiply of AB by R3-4, result in AB
*
?mul	MOV	R6,A		Save LO1
	MOV	A,R3		Get LO2
	MUL			Multiply HI1 * LO2
	MOV	R5,A		Save partial product
	MOV	B,R6		Get LO1
	MOV	A,R4		Get HI2
	MUL			Multiply LO1 * HI2
	MOV	R4,A		Save partial product
	MOV	A,R3		Get LO2
	MOV	B,R6		Get LO1
	MUL			Multiply LO2 * LO1
	XCH	A,B		Save low & get high
	ADD	A,R4		Add in carry from LO1 * HI2
	ADD	A,R5		Add in carry from HI1 * LO2
	XCH	A,B		Restore proper order
	RET
*
* 16 bit modulus of AB by R3-4
*
?mod	ACALL	?div		Perform the division
	MOV	B,R6		B = Remainder HIGH
	MOV	A,R5		A = Remainder LOW
	RET
*
* 16 bit division of AB % R3-4
*
?div	PUSH	?R2		Save index HIGH
	PUSH	?R1		Save index LOW
	MOV	R2,B		Save HI1
	MOV	R1,A		Save LO1
	MOV 	R6,#0		Begin with zero
	MOV 	R5,#0		"" ""
	MOV	R7,#17		16+1 bit shift
?div1	CLR	C		No carry in
?div2	MOV	A,R1		Shift parm1 left, C = carry out
	RLC	A		Do the shift
	MOV	R1,A		Resave
	MOV	A,R2		Get high half
	RLC	A		Do the shift
	MOV	R2,A		Resave
	DJNZ	R7,?div3	Not done, keep going
* Get result and exit
	MOV	B,A		B = Quotient HIGH
	MOV	A,R1		A = Quotient LOW
	POP	?R1		Restore index LOW
	POP	?R2		Restore index HIGH
	RET			And go home
* Compare registers & shift in approriate bit
?div3	MOV	A,R5		Shift carry into secondary result
	RLC	A		Do the shift
	MOV	R5,A		Resave
	MOV	A,R6		Get high half
	RLC	A		Do the shift
	MOV	R6,A		Resave
	CLR	C		Clear carry for test
	MOV	A,R5		Get secondary LOW
	SUBB	A,R3		Test against primary LOW
	MOV	B,A		Save for later store
	MOV	A,R6		Get secondary HIGH
	SUBB	A,R4		Test against poimary high
	JC	?div1		Still smaller, shift in 0
	MOV	R5,B		Set new secondary high
	MOV	R6,A		Set new secondary low
	SETB	C		Set '1' in register
	SJMP	?div2		Shift in one
* Shift accumulator LEFT by R3
?shl	CJNE	R3,#0,?shl1	Zero is special case
	RET			Perform no shift
?shl1	CLR	C		No shift in
	RLC	A		Shift low byte
	XCH	A,B		Get high
	RLC	A		Shift high byte
	XCH	A,B		Reorder ACC
	DJNZ	R3,?shl1	Repeat as necessary
	RET
* Shift accumulator RIGHT by R3
?shr	CJNE	R3,#0,?shr1	Zero is special case
	RET			Perform no shift
?shr1	CLR	C		No shift in
	XCH	A,B		Get high
	RRC	A		Shift high byte
	XCH	A,B		Get low
	RRC	A		Shift low byte
	DJNZ	R3,?shr1	Repeat as necessary
	RET
* Increment the 16 bit accumulator by 2
?inc2	ADD	A,#2		Advance by 2
	SJMP	*+4		And continue
* Increment the 16 bit accumulator by 1
?inc	ADD	A,#1		Increment ACC
	XCH	A,B		Swap in high
	ADDC	A,#0		Advance high if carry out
	XCH	A,B		Return positions
	RET
* Decrement the 16 bit accumulator by 2
?dec2	CLR	C		Begin with no "borrow"
	SUBB	A,#2		Reduce it by 2
	SJMP	*+5		And continue
* Decrement the 16 bit accumulator by 1
?dec	CLR	C		Begin with no "borrow"
	SUBB	A,#1		Redice it by 2
	XCH	A,B		Swap in high
	SUBB	A,#0		Decrement high if borrow
	XCH	A,B		Return position
	RET
* Two's complement the 16 bit accumulator
?neg	ACALL	?dec		Decrement the accumulator
* One's complement the 16 bit accumulator
?com	CPL	A		Complement low byte
	XCH	A,B		Swap it
	CPL	A		Complement high byte
	XCH	A,B		Reorder
	RET
* Logically negate AB
?not	ORL	A,B		Is it zero?
	JNZ	?ret0		No, makeit zero
	INC	A		Give a one
	RET
* Signed compares
?lt	ACALL	?scomp		AB < R3-4 (signed)
	JC	?ret1
	SJMP	?ret0
?le	ACALL	?scomp		AB <= R3-4 (signed)
	JC	?ret1
	JNZ	?ret0
	SJMP	?ret1
?gt	ACALL	?scomp		AB > R3-4 (signed)
	JC	?ret0
	JNZ	?ret1
	SJMP	?ret0
?ge	ACALL	?scomp		AB >= R3-4 (signed)
	JNC	?ret1
	SJMP	?ret0
* Equality compares
?eq	ACALL	?ucomp		AB == R3-4
	JZ	?ret1		They are the same
?ret0	CLR	A		Get zero
	MOV	B,A		Set high
	RET			Return logical ZERO
?ne	ACALL	?ucomp		AB != R3-4
	JZ	?ret0
?ret1	CLR	A		Get zero
	MOV	B,A		Set high
	INC	A		Convert to 1
	RET			Return logical ONE
* Unsigned compares
?ult	ACALL	?ucomp		AB < R3-4 (unsigned)
	JC	?ret1
	SJMP	?ret0
?ule	ACALL	?ucomp		AB <= R3-4 (unsigned)
	JC	?ret1
	JNZ	?ret0
	SJMP	?ret1
?ugt	ACALL	?ucomp		AB > R3-4 (unsigned)
	JC	?ret0
	JNZ	?ret1
	SJMP	?ret0
?uge	ACALL	?ucomp		AB >= R3-4 (unsigned)
	JNC	?ret1
	SJMP	?ret0
* Signed compare of AB & R3-4
?scomp	CLR	C		Clear borrow
	SUBB	A,R3		Calculate difference (low)
	XCH	A,B		Get high
	SUBB	A,R4		Calculate difference (high)
	JB	A.7,?sco2	Negative == <
	JB	PSW.2,?sco3	Overflow == <
?sco1	CLR	C		Indicate not less than
	ORL	A,B		Setup 'Z' if equal
	RET
?sco2	JB	PSW.2,?sco1	Negative & overflow == !<
?sco3	SETB	C		Indicate less than
?sco4	RET
* Unsigned compare of AB & R3-4
?ucomp	XCH	A,B		Get high
	CLR	C		Zero high
	SUBB	A,R4		Compare high
	JNZ	*+5		Different
	MOV	A,B		Get low
	SUBB	A,R3		Compare low
nargs	RET
*
* Locate a temporary buffer on the end of heap storage
*
?temp	MOV	DPTR,#?heap	Point to buffer
?temp1	MOVX	A,[DPTR]	Get flag value
	INC	DPTR		Skip flag
	JZ	?sco4		End of list
	MOVX	A,[DPTR]	Get LOW size
	MOV	R7,A		Save it
	INC	DPTR		Skip LOW
	MOVX	A,[DPTR]	Get HIGH size
	INC	DPTR		Skip HIGH
	XCH	A,R7		Get LOW
	ADD	A,DPL		Add to pointer
	MOV	DPL,A		Resave pointer
	MOV	A,R7		Get HIGH
	ADDC	A,DPH		Add to pointer
	MOV	DPH,A		Resave pointer
	SJMP	?temp1		And proceed
* Set up DPTR to access external stack
?accex1	MOV	R6,#0		Assume high byte of zero
?accex2	XCH	A,R7		Get LOW offset
	ADD	A,?stack	Add to stack LOW
	MOV	DPL,A		Save destination
	MOV	A,R6		Get HIGH Offset
	ADDC	A,?stack+1	Add to stack HIGH
	MOV	DPH,A		Save destination
	MOV	A,R7		Restore ACC
	RET
* Set up index register (R2:R1) to access external stack
?idxex1	MOV	R6,#0		Assume high byte of zero
?idxex2	XCH	A,R7		Get LOW offset
	ADD	A,?stack	Add to stack LOW
	MOV	R1,A		Set index LOW
	MOV	A,R6		Get HIGH offset
	ADDC	A,?stack+1	Add to stack HIGH
	MOV	R2,A		Save destination
	MOV	A,R7		Restore ACC
	RET
* Adjust external stack pointer by [R6:]R7 bytes
?adjex1	MOV	R6,#0		Assume high of zero
?adjex2	XCH	A,?stack	Get LOW SP
	ADD	A,R7		Advance
	XCH	A,?stack	Return it
	XCH	A,?stack+1	Get HIGH SP
	ADDC	A,R6		Include high
	XCH	A,?stack+1	And proceed
	RET
$SE:2
*
* Segment 2, internal "register" variables
*
	ORG	$0008		Internal ram ALWAYS starts here
?stack	DS	2		External stack pointer
