*
*----------------------------
* I2C BUSS Interface routines
*----------------------------
*
* Generate a START condition & TX char in A
*
?ICstrt	SETB	P3.4		SDA = 1
	SETB	P1.0		SCL = 1
	NOP
	CLR	P3.4		SDA = 0 - START
	NOP
	CLR	P1.0		SCL = 0 - Ready for first bit
*
* Send 8 bit character in A
*
?ICsend	MOV	R7,#8		8 bits ber byte
?1	JB	A.7,?2		Send a 1 bit
	CLR	P3.4		Send ZERO bit
	SJMP	?3
?2	SETB	P3.4		Send ONE bit
?3	SETB	P1.0		Toggle CLOCK high
	RL	A		Shift in next bit
	CLR	P1.0		Toggle CLOCK low
	DJNZ	R7,?1		Send all bits
	SETB	P3.4		Release DATA line
	SETB	P1.0		9'th clock
	NOP
	MOV	C,P3.4		Get ACK bit
	CLR	P1.0		Return to LOW
	RET
*
* Read 8 bits of data into ACC
*
?ICread	MOV	R7,#8		8 bits per byte
?4	SETB	P1.0		Set clock high
	RL	A		Shift it over
	MOV	C,P3.4		Get bit
	CLR	P1.0		Reset bit
	MOV	A.0,C		Set bit in RX
	DJNZ	R7,?4		Do them all
	RET
*
* Send an ACK to the remote
*
?ICack	CLR	P3.4		Zero DATA
	SETB	P1.0		Raise CLOCK
	NOP
	CLR	P1.0		Lower CLOCK
	SETB	P3.4		Release DATA
	RET
*
*-----------------------------------
* PCF8951 ADC/DAC Interface routines
*-----------------------------------
*
* Write DAC, Read all four ADC's: adc(value);
*
adc	MOV	R0,#-5		Address parm
	LCALL	?auto0		Address it
	MOV	A,#%10010010	ADC Write command
	LCALL	?ICstrt		Send it out
	JC	?6		Report error
	MOV	A,#%01000100	Enable DAC & AUTO INC
	LCALL	?ICsend		Send it out
	JC	?6		Report error
	MOV	A,[R0]		Get data
	LCALL	?ICsend		Send if out
	JC	?6		Report error
	LCALL	?ICstop		And terminate
* Read all ADC's
	MOV	A,#%10010011	ADC Read command
	LCALL	?ICstrt		Send it out
	JC	?6		Write failed
	LCALL	?ICread		Get data byte
	LCALL	?ICack		Ack first data
	LCALL	?ICread		Read Channel 0
	MOV	ADC0,A		R2 = Channel 0
	LCALL	?ICack		Ack second data
	LCALL	?ICread		Read Channel 1
	MOV	ADC1,A		R3 = Channel 1
	LCALL	?ICack		Ack third data
	LCALL	?ICread		Read channel 2
	MOV	ADC2,A		R4 = Channel 2
	LCALL	?ICack		Ack fourth data
	LCALL	?ICread		Read channel 3
	MOV	ADC3,A		R5 = channel 3
	SJMP	?8		And terminate
*
*--------------------------------------
* X2404/X2416 EEPROM Interface routines
*--------------------------------------
*
* Write data to EEPROM: eewrite(addr, data);
*
eewrite	MOV	R0,#-6		Point at parm
	LCALL	?auto0		Address it
	MOV	R6,#0		Set counter
?5	MOV	A,[R0]		Get HIGH address
	RL	A		Shift over
	ANL	A,#%00001110	Save only address
	ORL	A,#%10100000	Set for WRITE
	LCALL	?ICstrt		Start transfer
	JNC	?7		OK
	DJNZ	R6,?5		Keep re-trying
* Report error - return -1
?6	MOV	A,#-1		Get -1
	MOV	B,A		Copy to B
	SJMP	?ICstop		And exit
?7	DEC	R0		Backup to LOW
	MOV	A,[R0]		Get LOW address
	LCALL	?ICsend		Write to chip
	JC	?6		Report error
	INC	R0		Skip to HIGH
	INC	R0		Skip to DATA
	MOV	A,[R0]		Get data back
	LCALL	?ICsend		Write to chip
	JC	?6		Report error
* Return zero
?8	CLR	A		Zero LOW
	MOV	B,A		Zero HIGH
*
* Send the STOP condition to the remote
*
?ICstop	CLR	P3.4		Zero DATA
	SETB	P1.0		Raise CLOCK
	NOP
	SETB	P3.4		Raise DATA
	RET
*
* Read data from EEPROM: eeread(addr);
*
eeread	MOV	R0,#-4		Access first
	LCALL	?auto0		Access stack
	MOV	R6,#0		Zero retry count
?9	MOV	A,[R0]		Get HIGH address
	RL	A		Shift over
	ANL	A,#%00001110	Save only address
	ORL	A,#%10100000	Set for WRITE
	LCALL	?ICstrt		Start transfer
	JNC	?10		Its OK
	DJNZ	R6,?9		Do full retry
	SJMP	?6		Report error
?10	DEC	R0		Backup to LOW
	MOV	A,[R0]		Get LOW address
	LCALL	?ICsend		Write
	JC	?6		Report error
	MOV	A,R1		Get HIGH address
	RL	A		Shift over
	ANL	A,#%00001110	Save only address
	ORL	A,#%10100001	Set for READ
	LCALL	?ICstrt		ReStart transfer
	JC	?6		Report error
	LCALL	?ICread		Get data byte
	MOV	B,#0		Zero high
	SJMP	?ICstop		Stop transfer
$SE:2
* Analog to Digital convertor results: extern unsigned char ADCx;
ADC0	DS	1		ADC0 results
ADC1	DS	1		ADC1 results
ADC2	DS	1		ADC2 results
ADC3	DS	1		ADC3 results
