*
* Allocate a block of memory
*
malloc	MOV	R0,#-5		Argument address
	LCALL	?auto0		Point to it
* Search for a free block of memory
	MOV	DPTR,#?heap	Point to heap
?1	MOV	R6,DPH		Save HIGH
	MOV	R5,DPL		Save LOW
	MOVX	A,[DPTR]	Get memory flag
	MOV	R2,A		Save type
	JZ	?2		End of list, allocate here
	LCALL	?30		Advance to next block
	DJNZ	R2,?1		Not free get next
* Test this block to see if it is large enough
	MOV	A,[R0]		Get LOW requested size
	INC	R0		Advance
	MOV	B,[R0]		Get HIGH requested SIZE
	DEC	R0		Backup
	LCALL	?gt		Request > block?
	JNZ	?1		Yes, go again
* This block is OK to re-allocate
* First calculate the remainder after allocating the block
	MOV	A,R3		Get LOW block size
	CLR	C		Insure no carry
	SUBB	A,[R0]		Calculate remainder
	MOV	R3,A		Save result
	INC	R0		Advance register
	MOV	A,R4		Get HIGH block size
	SUBB	A,[R0]		Calculate high
* Adjust remainder for overhead size
	XCH	A,R3		Swap into position
	SUBB	A,#3		Adjust for position
	XCH	A,R3		Save LOW result
	SUBB	A,#0		Adjust for position
	MOV	R4,A		Save HIGH result
* If large enough, block can be split
	JC	?4		Too small
	ORL	A,R3		Zero result?
	JZ	?4		Too small
* Split this block into two blocks
	DEC	R0		Backup to LOW size
	LCALL	?40		Point to end
	MOV	A,#1		Get de-allocated flag
	MOVX	[DPTR],A	Write to structure
	INC	DPTR		Advance to next
	MOV	A,R3		Get LOW size
	MOVX	[DPTR],A	Write to structure
	INC	DPTR		Advance
	MOV	A,R4		Get HIGH size
	MOVX	[DPTR],A	Write to structure
	SJMP	?3		Set the size
* Allocate on the end of memory
?2	LCALL	?40		Point to end
	CLR	A		End of list flag
	MOVX	[DPTR],A	Set it
* Set new size for this block
?3	MOV	DPH,R6		Restore pointer HIGH
	MOV	DPL,R5		Restore pointer LOW
	INC	DPTR		Skip flag
	DEC	R0		Backup to LOW size
	MOV	A,[R0]		Get LOW size
	MOVX	[DPTR],A	Write to structure
	INC	DPTR		Advance
	INC	R0		Advance to HIGH size
	MOV	A,[R0]		Get HIGH size
	MOVX	[DPTR],A	Write to structure
* Mark this block as "allocated"
?4	MOV	DPH,R6		Restore pointer
	MOV	DPL,R5		To block
	MOV	A,#2		Get SET flag
	MOVX	[DPTR],A	Set it
	INC	DPTR		Skip flag
	INC	DPTR		Skip LOW size
	INC	DPTR		Skip HIGH size
	MOV	B,DPH		Get HIGH address
	MOV	A,DPL		Get LOW address
	RET
*
* Release a block of memory
*
free	MOV	R0,#-5		Point to parm
	LCALL	?auto0		Address it
	MOV	A,[R0]		Get LOW parm
	INC	R0		Advance to high
	CLR	C		No carry
	SUBB	A,#3		Backup
	MOV	R5,A		Save LOW
	MOV	A,[R0]		Get HIGH
	SUBB	A,#0		Adjust high
	MOV	R6,A		Save HIGH
	MOV	DPTR,#?heap	Point to heap
?10	MOVX	A,[DPTR]	Get list pointer
	JZ	?20		End, collect garbage
	MOV	A,R5		Get LOW
	CJNE	A,DPL,?11	No match
	MOV	A,R6		Get HIGH
	CJNE	A,DPH,?11	No match
	MOV	A,#1		Get released flag
	MOVX	[DPTR],A	Set it
	SJMP	?20		Collect garbage
?11	LCALL	?30		Advance to next block
	SJMP	?10		And try another
*
* Garbage collection
*
?20	MOV	DPTR,#?heap	Point to heap memory
?21	MOV	R6,DPH		Save pointer HIGH
	MOV	R5,DPL		Save pointer LOW
	MOVX	A,[DPTR]	Get flag byte
	MOV	R1,A		Save the flag
	JZ	?25		Yes, quit
	LCALL	?30		Advance to next	
	DJNZ	R1,?21		Not free, keep looking
	MOV	R2,?R3		Save size LOW
	MOV	R7,?R4		Save size HIGH
* Block is free, keep looking at following blocks
?22	MOVX	A,[DPTR]	Get block flag?
	MOV	R1,A		Save it
	JZ	?24		End of list
	LCALL	?30		Skip to next
	DJNZ	R1,?23		This block not free
* Include this size (+ overhead) into running total
	MOV	A,R2		Get total LOW
	ADD	A,R3		Add size LOW
	MOV	R2,A		Resave
	MOV	A,R7		Get total HIGH
	ADDC	A,R4		Add size HIGH
	MOV	R7,A		Resave
	MOV	A,R2		Get total LOW
	ADD	A,#3		Add adjust LOW
	MOV	R2,A		Resave
	MOV	A,R7		Get total HIGH
	ADDC	A,#0		Adjust high
	MOV	R7,A		Resave
	SJMP	?22		Keep looking
* Resave this block size
?23	MOV	DPH,R6		Restore pointer HIGH
	MOV	DPL,R5		Restore pointer LOW
	INC	DPTR		Advance
	MOV	A,R2		Get total size LOW
	MOVX	[DPTR],A	Save total size
	INC	DPTR		Advance
	MOV	A,R7		Get total size HIGH
	MOVX	[DPTR],A	Save total size
	MOV	DPH,R6		Restore pointer HIGH
	MOV	DPL,R5		Restore pointer LOW
	LCALL	?30		Advance to next
	SJMP	?21		And proceed
?24	MOV	DPH,R6		Restore pointer HIGH
	MOV	DPL,R5		Restore pointer LOW
	CLR	A		End of list flag
	MOVX	[DPTR],A	Set block flag
?25	RET
*
* Advance to next block & keep size
*
?30	INC	DPTR
	MOVX	A,[DPTR]	Get size high
	MOV	R3,A		Save size LOW
	INC	DPTR		Advance to size HIGH
	MOVX	A,[DPTR]	Get size high
	MOV	R4,A		Save size HIGH
	INC	DPTR		Advance to data
	MOV	A,DPL		Get LOW address
	ADD	A,R3		Adjust
	MOV	DPL,A		Save adjust
	MOV	A,DPH		Get HIGH address
	ADDC	A,R4		Adjust
	MOV	DPH,A		Save it
	RET
*
* Set DPTR to point to end of allocated block
*
?40	MOV	A,R5		Get LOW address
	ADD	A,[R0]		Adjust
	INC	R0		Advance to size H
	MOV	DPL,A		Save LOW pointer
	MOV	A,R6		Get HIGH address
	ADDC	A,[R0]		Adjust
	MOV	DPH,A		Save HIGH pointer
	INC	DPTR		Skip flag
	INC	DPTR		Skip size LOW
	INC	DPTR		Skip size HIGH
	RET
